#include <pxp-agent/util/posix/pid_file.hpp>
#include <pxp-agent/util/process.hpp>

#include <boost/filesystem.hpp>
#include <boost/filesystem/operations.hpp>

#include <leatherman/file_util/file.hpp>

#include <leatherman/locale/locale.hpp>

#define LEATHERMAN_LOGGING_NAMESPACE "puppetlabs.pxp_agent.util.posix.pid_file"
#include <leatherman/logging/logging.hpp>

#include <cstdio>           // remove()
#include <cassert>
#include <sys/file.h>       // open()
#include <sys/stat.h>
#include <fcntl.h>          // fcntl(), open() flags
#include <signal.h>
#include <unistd.h>         // getpid()

namespace PXPAgent {
namespace Util {

namespace fs = boost::filesystem;
namespace lth_file = leatherman::file_util;
namespace lth_loc  = leatherman::locale;

PIDFile::PIDFile(const std::string& file_path_)
        : dir_path { fs::path(file_path_).parent_path().string() },
          file_path { file_path_ },
          pidfile_fd {},
          cleanup_when_done { false } {
    pidfile_fd = open(file_path.data(), O_RDWR | O_CREAT, 0640);

    if (pidfile_fd == -1)
        throw PIDFile::Error {
            lth_loc::format("failed to open PID file '{1}'; {2} ({3})",
                            file_path, strerror(errno), errno) };
}

PIDFile::~PIDFile() {
    if (cleanup_when_done) {
        try {
            cleanup();
        } catch (const std::exception& e) {
            LOG_ERROR("Failed to clean up PID file '{1}': {2}",
                      file_path, e.what());
        }
    }
}

bool PIDFile::isExecuting() {
    try {
        auto pid = PIDFile::read();
        return processExists(pid);
    } catch (const PIDFile::Error& e) {
        // NB: processExists() does not throw
        LOG_DEBUG("Couldn't retrieve PID from file '{1}': {2}",
                  file_path, e.what());
    }

    return false;
}

void PIDFile::lockRead(bool blocking) {
    if (blocking){
        lockFileBlocking(pidfile_fd, F_RDLCK);
    } else {
        lockFile(pidfile_fd, F_RDLCK);
    }
}

void PIDFile::lockWrite(bool blocking) {
    if (blocking){
        lockFileBlocking(pidfile_fd, F_WRLCK);
    } else {
        lockFile(pidfile_fd, F_WRLCK);
    }
}

bool PIDFile::canLockRead() {
    return canLockFile(pidfile_fd, F_RDLCK);
}

bool PIDFile::canLockWrite() {
    return canLockFile(pidfile_fd, F_WRLCK);
}

void PIDFile::write(const pid_t& pid) {
    try {
        lth_file::atomic_write_to_file(std::to_string(pid) + "\n", file_path);
    } catch (const std::exception& e) {
        throw PIDFile::Error {
            lth_loc::format("failed to write '{1}': {2}", file_path, e.what()) };
    }
}

pid_t PIDFile::read() {
    if (!fs::exists(file_path)) {
        throw PIDFile::Error { lth_loc::translate("PID file does not exist") };
    }

    if (!fs::is_regular_file(file_path)) {
        throw PIDFile::Error {
            lth_loc::translate("PID file is not a regular file") };
    }

    if (!lth_file::file_readable(file_path)) {
        throw PIDFile::Error { lth_loc::translate("cannot read PID file") };
    }

    auto pid_txt = lth_file::read(file_path);

    if (pid_txt.empty()) {
        throw PIDFile::Error { lth_loc::translate("PID file is empty") };
    }

    try {
        auto pid = std::stoi(pid_txt);
        return static_cast<pid_t>(pid);
    } catch (const std::invalid_argument& e) {
        throw PIDFile::Error {
            lth_loc::translate("invalid value stored in PID file") };
    }
}

void PIDFile::cleanup() {
    LOG_DEBUG("Unlocking PID file {1} and closing its open file descriptor",
              file_path);
    PIDFile::unlockFile(pidfile_fd);
    close(pidfile_fd);

    if (fs::exists(file_path) && fs::is_regular_file(file_path)) {
        if (lth_file::file_readable(file_path)) {
            if (std::remove(file_path.data())) {
                LOG_ERROR("Failed to remove PID file '{1}'", file_path);
            } else {
                LOG_DEBUG("Removed PID file '{1}'", file_path);
            }
        } else {
            LOG_DEBUG("Cannot access PID file '{1}'", file_path);
        }
    } else {
        LOG_DEBUG("PID file '{1}' does not exist", file_path);
    }
}

void PIDFile::cleanupWhenDone() {
    cleanup_when_done = true;
}

// Static function members

static int callFcntl(int fd, int cmd, int lock_type) {
    struct flock fl;
    fl.l_type = lock_type;
    fl.l_whence = SEEK_SET;
    fl.l_start = 0;
    fl.l_len = 0;

    return fcntl(fd, cmd, &fl);
}

void PIDFile::lockFile(int fd, int lock_type) {
    // NOTE(ale): fcntl-based lock files are process / inode specific
    assert(lock_type == F_RDLCK || lock_type == F_WRLCK);

    if (callFcntl(fd, F_SETLK, lock_type) == -1) {
        if (errno == EAGAIN || errno == EACCES) {
            throw PIDFile::Error {
                lth_loc::translate("incompatible with an existent lock") };
        } else if (errno == EDEADLK) {
            throw PIDFile::Error { lth_loc::translate("deadlock") };
        } else {
            // Unexpected; see fcntl's man 2
            throw PIDFile::Error {
                lth_loc::format("unexpected error with errno={1}", errno) };
        }
    }
}

void PIDFile::lockFileBlocking(int fd, int lock_type) {
    assert(lock_type == F_RDLCK || lock_type == F_WRLCK);

    if (callFcntl(fd, F_SETLKW, lock_type) == -1) {
        if (errno == EINTR) {
            throw PIDFile::Error {
                lth_loc::translate("caught an interrupt signal") };
        } else {
            // Unexpected; see fcntl's man 2
            throw PIDFile::Error {
                lth_loc::format("unexpected error with errno={1}", errno) };
        }
    }
}

void PIDFile::unlockFile(int fd) {
    // NOTE(ale): it is always safe to unlock with fcntl()
    if (callFcntl(fd, F_SETLK, F_UNLCK) == -1)
        throw PIDFile::Error { lth_loc::format("errno={1}", errno) };
}

bool PIDFile::canLockFile(int fd, int lock_type) {
    assert(lock_type == F_RDLCK || lock_type == F_WRLCK);

    struct flock fl;
    fl.l_type = lock_type;
    fl.l_whence = SEEK_SET;
    fl.l_start = 0;
    fl.l_len = 0;

    if (fcntl(fd, F_GETLK, &fl) == -1)
        // Failure
        throw PIDFile::Error { lth_loc::format("errno={1}", errno) };

    return (fl.l_type == F_UNLCK);
}

}  // namespace Util
}  // namespace PXPAgent
